@@ -3,20 +3,23 @@
 import datetime
 from collections import deque, defaultdict
 from heapq import heappop, heappush
-from trytond.model import ModelView, ModelSQL, fields
-from trytond.wizard import Wizard
+
+from trytond.model import ModelSQL, fields
+from trytond.wizard import Wizard, StateTransition
 from trytond.pyson import Eval
 from trytond.transaction import Transaction
-from trytond.pool import Pool
+from trytond.pool import Pool, PoolMeta
+
+__all__ = ['Work', 'PredecessorSuccessor', 'Leveling']
+__metaclass__ = PoolMeta
 
 
 def intfloor(x):
     return int(round(x, 4))
 
 
-class Work(ModelSQL, ModelView):
-    _name = 'project.work'
-
+class Work:
+    __name__ = 'project.work'
     predecessors = fields.Many2Many('project.predecessor_successor',
         'successor', 'predecessor', 'Predecessors',
         domain=[
@@ -31,8 +34,8 @@ class Work(ModelSQL, ModelView):
             ('id', '!=', Eval('id')),
             ],
         depends=['parent', 'id'])
-    leveling_delay = fields.Float("Leveling Delay")
-    back_leveling_delay = fields.Float("Leveling Delay")
+    leveling_delay = fields.Float("Leveling Delay", required=True)
+    back_leveling_delay = fields.Float("Back Leveling Delay", required=True)
     allocations = fields.One2Many('project.allocation', 'work', 'Allocations',
         states={
             'invisible': Eval('type') != 'task',
@@ -64,93 +67,72 @@ class Work(ModelSQL, ModelView):
     constraint_finish_date = fields.Function(fields.Date('Constraint Finish',
         depends=['type']), 'get_function_fields',
         setter='set_function_fields')
-    requests = fields.Function(fields.One2Many('res.request', None,
-        'Requests'), 'get_function_fields', setter='set_function_fields')
 
-    def __init__(self):
-        super(Work, self).__init__()
-        self._constraints += [
-            ('check_recursion', 'recursive_dependency'),
-            ]
-        self._error_messages.update({
-                'recursive_dependency': 'You can not create recursive '
-                        'dependencies!',
-                })
+    @classmethod
+    def __setup__(cls):
+        super(Work, cls).__setup__()
 
-    def check_recursion(self, ids, parent='parent'):
-        return super(Work, self).check_recursion(ids, parent='successors')
+    @classmethod
+    def validate(cls, works):
+        super(Work, cls).validate(works)
+        cls.check_recursion(works, parent='successors')
 
-    def get_function_fields(self, ids, names):
+    @staticmethod
+    def default_leveling_delay():
+        return 0.0
+
+    @staticmethod
+    def default_back_leveling_delay():
+        return 0.0
+
+    @classmethod
+    def get_function_fields(cls, works, names):
         '''
         Function to compute function fields
-
-        :param ids: the ids of the works
-        :param names: the list of field name to compute
-        :return: a dictionary with all field names as key and
-            a dictionary as value with id as key
         '''
-        req_ref_obj = Pool().get('res.request.reference')
-
-        cursor = Transaction().cursor
-
         res = {}
 
-        if 'requests' in names:
-            requests = dict((i, []) for i in ids)
-
-            for i in range(0, len(ids), cursor.IN_MAX):
-                sub_ids = ids[i:i + cursor.IN_MAX]
-
-                req_ref_ids = req_ref_obj.search([
-                        ('reference', 'in', [
-                                'project.work,%s' % i for i in sub_ids
-                                ]
-                         ),
-                        ])
-                req_refs = req_ref_obj.browse(req_ref_ids)
-                for req_ref in req_refs:
-                    _, work_id = req_ref.reference.split(',')
-                    requests[int(work_id)].append(req_ref.request.id)
-
-            res['requests'] = requests
-
+        ids = [w.id for w in works]
         if 'duration' in names:
-            all_ids = self.search([
+            all_works = cls.search([
                     ('parent', 'child_of', ids),
-                    ('active', '=', True)]) + ids
-            all_ids = list(set(all_ids))
-
-            works = self.browse(all_ids)
+                    ('active', '=', True)]) + works
+            all_works = set(all_works)
 
             durations = {}
             id2work = {}
             leafs = set()
-            for work in works:
+            for work in all_works:
                 id2work[work.id] = work
                 if not work.children:
                     leafs.add(work.id)
 
                 total_allocation = 0
                 if not work.allocations:
-                    durations[work.id] = work.effort
+                    durations[work.id] = (work.effort or 0)
                     continue
                 for allocation in work.allocations:
                     total_allocation += allocation.percentage
-                durations[work.id] = work.effort / (total_allocation / 100.0)
+                durations[work.id] = ((work.effort or 0)
+                    / (total_allocation / 100.0))
 
             while leafs:
-                parents = set()
                 for work_id in leafs:
                     work = id2work[work_id]
+                    all_works.remove(work)
                     if not work.active:
                         continue
                     if work.parent and work.parent.id in durations:
                         durations[work.parent.id] += durations[work_id]
-                        parents.add(work.parent.id)
-                leafs = parents
+                next_leafs = set(w.id for w in all_works)
+                for work in all_works:
+                    if not work.parent:
+                        continue
+                    if work.parent.id in next_leafs and work.parent in works:
+                        next_leafs.remove(work.parent.id)
+                leafs = next_leafs
             res['duration'] = durations
 
-
         fun_fields = ('early_start_date', 'early_finish_date',
                       'late_start_date', 'late_finish_date',
                       'actual_start_date', 'actual_finish_date',
@@ -160,103 +142,40 @@ class Work(ModelSQL, ModelView):
                   'actual_start_time', 'actual_finish_time',
                   'constraint_start_time', 'constraint_finish_time')
 
-        works = None
         for fun_field, db_field in zip(fun_fields, db_fields):
             if fun_field in names:
                 values = {}
-                if works is None:
-                    works = self.browse(ids)
                 for work in works:
-                    values[work.id] = work[db_field] \
-                        and work[db_field].date() or None
+                    values[work.id] = getattr(work, db_field) \
+                        and getattr(work, db_field).date() or None
                 res[fun_field] = values
 
         return res
 
-    def set_function_fields(self, ids, name, value):
-        request_obj = Pool().get('res.request')
-        req_ref_obj = Pool().get('res.request.reference')
-
-        if name == 'requests':
-            works = self.browse(ids)
-            currents = dict((req.id, req) for work in works for req in
-                    work.requests)
-            if not value:
-                return
-            for v in value:
-                to_unlink = []
-                to_link = []
-                operator = v[0]
-
-                target_ids = len(v) > 1 and v[1] or []
-                if operator == 'create':
-                    to_link.append(request_obj.create(v[1]))
-                elif operator == 'write':
-                    request_obj.write(v[1], v[2])
-                elif operator == 'delete':
-                    request_obj.delete(v[1])
-                elif operator == 'delete_all':
-                    target_ids = []
-                    for record_id in ids:
-                        ref_ids = req_ref_obj.search([
-                                ('reference', '=',
-                                    'project.work,%s' % record_id),
-                                ])
-                        refs = req_ref_obj.browse(ref_ids)
-                        target_ids.extend(ref.request.id for ref in refs)
-                    request_obj.delete(target_ids)
-                elif operator == 'unlink':
-                    to_unlink.extend((i for i in target_ids if i in currents))
-                elif operator == 'add':
-                    to_link.extend((i for i in target_ids if i not in
-                        currents))
-                elif operator == 'unlink_all':
-                    to_unlink.extend(currents)
-                elif operator == 'set':
-                    to_link.extend((i for i in target_ids
-                        if i not in currents))
-                    to_unlink.extend((i for i in currents
-                        if i not in target_ids))
-                else:
-                    raise Exception('Operation not supported')
-
-                req_ref_ids = []
-                for i in to_unlink:
-                    request = currents[i]
-                    for ref in request.references:
-                        if int(ref.reference.split(',')[1]) in ids:
-                            req_ref_ids.append(ref.id)
-                req_ref_obj.delete(req_ref_ids)
-
-                for i in to_link:
-                    for record_id in ids:
-                        req_ref_obj.create({
-                                'request': i,
-                                'reference': 'project.work,%s' % record_id,
-                                })
-            return
-
+    @classmethod
+    def set_function_fields(cls, works, name, value):
         fun_fields = ('actual_start_date', 'actual_finish_date',
                       'constraint_start_date', 'constraint_finish_date')
         db_fields = ('actual_start_time', 'actual_finish_time',
                      'constraint_start_time', 'constraint_finish_time')
         for fun_field, db_field in zip(fun_fields, db_fields):
             if fun_field == name:
-                self.write(ids, {
-                        db_field: value \
-                                and datetime.datetime.combine(value,
-                                    datetime.time()) \
-                                or False,
+                cls.write(works, {
+                        db_field: (value
+                            and datetime.datetime.combine(value,
+                                datetime.time())
+                            or None),
                         })
                 break
 
-    def add_minutes(self, company, date, minutes):
+    @classmethod
+    def add_minutes(cls, company, date, minutes):
         minutes = int(round(minutes))
         minutes = date.minute + minutes
 
         hours = minutes // 60
         if hours:
-            date = self.add_hours(company, date, hours)
+            date = cls.add_hours(company, date, hours)
 
         minutes = minutes % 60
 
@@ -270,19 +189,18 @@ class Work(ModelSQL, ModelView):
 
         return date
 
-    def add_hours(self, company, date, hours):
-        day_per_week = company.hours_per_work_week / company.hours_per_work_day
-
+    @classmethod
+    def add_hours(cls, company, date, hours):
         while hours:
-            if hours !=  intfloor(hours):
+            if hours != intfloor(hours):
                 minutes = (hours - intfloor(hours)) * 60
-                date = self.add_minutes(company, date, minutes)
+                date = cls.add_minutes(company, date, minutes)
             hours = intfloor(hours)
 
             hours = date.hour + hours
             days = hours // company.hours_per_work_day
             if days:
-                date = self.add_days(company, date, days)
+                date = cls.add_days(company, date, days)
 
             hours = hours % company.hours_per_work_day
 
@@ -298,13 +216,14 @@ class Work(ModelSQL, ModelView):
 
         return date
 
-    def add_days(self, company, date, days):
+    @classmethod
+    def add_days(cls, company, date, days):
         day_per_week = company.hours_per_work_week / company.hours_per_work_day
 
         while days:
-            if days !=  intfloor(days):
+            if days != intfloor(days):
                 hours = (days - intfloor(days)) * company.hours_per_work_day
-                date = self.add_hours(company, date, hours)
+                date = cls.add_hours(company, date, hours)
             days = intfloor(days)
 
             days = date.weekday() + days
@@ -313,48 +232,47 @@ class Work(ModelSQL, ModelView):
             days = days % day_per_week
 
             if weeks:
-                date = self.add_weeks(company, date, weeks)
+                date = cls.add_weeks(company, date, weeks)
 
-            date += datetime.timedelta(days= -date.weekday() + intfloor(days))
+            date += datetime.timedelta(days=-date.weekday() + intfloor(days))
 
             days = days - intfloor(days)
 
         return date
 
-    def add_weeks(self, company, date, weeks):
+    @classmethod
+    def add_weeks(cls, company, date, weeks):
         day_per_week = company.hours_per_work_week / company.hours_per_work_day
 
-        if weeks !=  intfloor(weeks):
+        if weeks != intfloor(weeks):
             days = (weeks - intfloor(weeks)) * day_per_week
             if days:
-                date = self.add_days(company, date, days)
+                date = cls.add_days(company, date, days)
 
-        date += datetime.timedelta(days= 7 * intfloor(weeks))
+        date += datetime.timedelta(days=7 * intfloor(weeks))
 
         return date
 
-    def compute_dates(self, work_id):
-        active_work = self.browse(work_id)
+    def compute_dates(self):
         values = {}
         get_early_finish = lambda work: values.get(work, {}).get(
-            'early_finish_time', work['early_finish_time'])
+            'early_finish_time', work.early_finish_time)
         get_late_start = lambda work: values.get(work, {}).get(
-            'late_start_time', work['late_start_time'])
-        maxdate = lambda x,y: x and y and max(x,y) or x or y
-        mindate = lambda x,y: x and y and min(x,y) or x or y
-
+            'late_start_time', work.late_start_time)
+        maxdate = lambda x, y: x and y and max(x, y) or x or y
+        mindate = lambda x, y: x and y and min(x, y) or x or y
 
         # propagate constraint_start_time
-        constraint_start = reduce(maxdate, (pred.early_finish_time \
-                for pred in active_work.predecessors), None)
+        constraint_start = reduce(maxdate, (pred.early_finish_time
+                for pred in self.predecessors), None)
 
-        if constraint_start is None and active_work.parent:
-            constraint_start = active_work.parent.early_start_time
+        if constraint_start is None and self.parent:
+            constraint_start = self.parent.early_start_time
 
         constraint_start = maxdate(constraint_start,
-                                   active_work.constraint_start_time)
+            self.constraint_start_time)
 
-        works = deque([(active_work, constraint_start)])
+        works = deque([(self, constraint_start)])
         work2children = {}
         parent = None
 
@@ -426,16 +344,16 @@ class Work(ModelSQL, ModelView):
                 parent = work
 
         # propagate constraint_finish_time
-        constraint_finish = reduce(mindate, (succ.late_start_time \
-                for succ in active_work.successors), None)
+        constraint_finish = reduce(mindate, (succ.late_start_time
+                for succ in self.successors), None)
 
-        if constraint_finish is None and active_work.parent:
-            constraint_finish = active_work.parent.late_finish_time
+        if constraint_finish is None and self.parent:
+            constraint_finish = self.parent.late_finish_time
 
         constraint_finish = mindate(constraint_finish,
-                                    active_work.constraint_finish_time)
+            self.constraint_finish_time)
 
-        works = deque([(active_work, constraint_finish)])
+        works = deque([(self, constraint_finish)])
         work2children = {}
         parent = None
 
@@ -511,31 +429,29 @@ class Work(ModelSQL, ModelView):
         for work, val in values.iteritems():
             write_cond = False
             for field in write_fields:
-                if field in val and work[field] != val[field]:
+                if field in val and getattr(work, field) != val[field]:
                     write_cond = True
                     break
 
             if write_cond:
-                self.write(work.id, val)
+                self.write([work], val)
 
-    def reset_leveling(self, work_id):
+    def reset_leveling(self):
         get_key = lambda w: (set(p.id for p in w.predecessors),
                              set(s.id for s in w.successors))
 
-        work = self.browse(work_id)
-        parent_id = work.parent and work.parent.id or False
-        sibling_ids = self.search([
+        parent_id = self.parent and self.parent.id or None
+        siblings = self.search([
                 ('parent', '=', parent_id)
                 ])
-        siblings = self.browse(sibling_ids)
         to_clean = []
 
-        ref_key = get_key(work)
+        ref_key = get_key(self)
         for sibling in siblings:
             if sibling.leveling_delay == sibling.back_leveling_delay == 0:
                 continue
             if get_key(sibling) == ref_key:
-                to_clean.append(sibling.id)
+                to_clean.append(sibling)
 
         if to_clean:
             self.write(to_clean, {
@@ -543,15 +459,17 @@ class Work(ModelSQL, ModelView):
                     'back_leveling_delay': 0,
                     })
 
-    def create_leveling(self, work_id):
+    def create_leveling(self):
         # define some helper functions
         get_key = lambda w: (set(p.id for p in w.predecessors),
                              set(s.id for s in w.successors))
-        over_alloc = lambda current_alloc, work: \
-            reduce(lambda res, alloc: res or current_alloc[alloc.employee.id] + \
-                       alloc.percentage > 100,
-                   work.allocations,
-                   False)
+        over_alloc = lambda current_alloc, work: (
+            reduce(lambda res, alloc: (res
+                    or (current_alloc[alloc.employee.id]
+                        + alloc.percentage) > 100),
+                work.allocations,
+                False))
+
         def sum_allocs(current_alloc, work):
             res = defaultdict(float)
             for alloc in work.allocations:
@@ -597,148 +515,126 @@ class Work(ModelSQL, ModelView):
 
                 yield sibling, delay
 
-        work = self.browse(work_id)
-        parent_id = work.parent and work.parent.id or False
-        sibling_ids = self.search([
-                ('parent', '=', parent_id)
+        siblings = self.search([
+                ('parent', '=', self.parent.id if self.parent else None)
                 ])
 
-        refkey = get_key(work)
-        siblings = [s for s in self.browse(sibling_ids) if get_key(s) == refkey]
+        refkey = get_key(self)
+        siblings = [s for s in siblings if get_key(s) == refkey]
 
         for sibling, delay in compute_delays(siblings):
-            self.write(sibling.id, {
+            self.write([sibling], {
                     'leveling_delay': delay,
                     })
 
         siblings.reverse()
         for sibling, delay in compute_delays(siblings):
-            self.write(sibling.id, {
+            self.write([sibling], {
                     'back_leveling_delay': delay,
                     })
 
-        if parent_id:
-            self.compute_dates(parent_id)
+        if self.parent:
+            self.parent.compute_dates()
 
-    def write(self, ids, values):
-        res = super(Work, self).write(ids, values)
-        if isinstance(ids, (int, long)):
-            ids = [ids]
+    @classmethod
+    def write(cls, *args):
+        super(Work, cls).write(*args)
 
-        if 'effort' in values:
-            for work_id in ids:
-                self.reset_leveling(work_id)
-        fields = ('constraint_start_time', 'constraint_finish_time',
-                  'effort')
-        if reduce(lambda x, y : x or y in values, fields, False):
-            for work_id in ids:
-                self.compute_dates(work_id)
-        return res
+        actions = iter(args)
+        for works, values in zip(actions, actions):
+            if 'effort' in values:
+                for work in works:
+                    work.reset_leveling()
+            fields = ('constraint_start_time', 'constraint_finish_time',
+                      'effort')
+            if reduce(lambda x, y: x or y in values, fields, False):
+                for work in works:
+                    work.compute_dates()
 
-    def create(self, values):
-        work_id = super(Work, self).create(values)
-        self.reset_leveling(work_id)
-        self.compute_dates(work_id)
-        return work_id
+    @classmethod
+    def create(cls, vlist):
+        works = super(Work, cls).create(vlist)
+        for work in works:
+            work.reset_leveling()
+            work.compute_dates()
+        return works
 
-    def delete(self, ids):
-        if isinstance(ids, (int, long)):
-            ids = [ids]
-        works = self.browse(ids)
+    @classmethod
+    def delete(cls, works):
         to_update = set()
         for work in works:
-            if work.parent and work.parent.id not in ids:
-                to_update.add(work.parent.id)
-                to_update.update(c.id for c in work.parent.children \
-                                     if c.id not in ids)
-        res = super(Work, self).delete(ids)
-
-        for work_id in to_update:
-            self.reset_leveling(work_id)
-            self.compute_dates(work_id)
-
-        return res
-Work()
+            if work.parent and work.parent not in works:
+                to_update.add(work.parent)
+                to_update.update(c for c in work.parent.children
+                    if c not in works)
+        super(Work, cls).delete(works)
 
+        for work in to_update:
+            work.reset_leveling()
+            work.compute_dates()
 
 
 class PredecessorSuccessor(ModelSQL):
     'Predecessor - Successor'
-    _name = 'project.predecessor_successor'
-    _description = __doc__
-
+    __name__ = 'project.predecessor_successor'
     predecessor = fields.Many2One('project.work', 'Predecessor',
-            ondelete='CASCADE', required=True, select=1)
+            ondelete='CASCADE', required=True, select=True)
     successor = fields.Many2One('project.work', 'Successor',
-            ondelete='CASCADE', required=True, select=1)
+            ondelete='CASCADE', required=True, select=True)
 
-    def write(self, ids, values):
-        work_obj = Pool().get('project.work')
-        res = super(PredecessorSuccessor, self).write(ids, values)
+    @classmethod
+    def write(cls, *args):
+        Work = Pool().get('project.work')
+        super(PredecessorSuccessor, cls).write(*args)
 
-        for work_id in values.itervalues():
-            work_obj.reset_leveling(work_id)
-        for work_id in values.itervalues():
-            work_obj.compute_dates(work_id)
-        return res
+        work_ids = [v for values in args[1::2]
+            for k, v in values.itervalues()
+            if k in ('predecessor', 'successor')]
+        works = Work.browse(work_ids)
+        for work in works:
+            work.reset_leveling()
+        for work in works:
+            work.compute_dates()
 
-    def delete(self, ids):
-        work_obj = Pool().get('project.work')
-        if isinstance(ids, (int, long)):
-            ids = [ids]
-
-        work_ids = set()
-        parent_ids = set()
-        pred_succs = self.browse(ids)
+    @classmethod
+    def delete(cls, pred_succs):
+        works = set()
+        parents = set()
         for pred_succ in pred_succs:
-            work_ids.update((pred_succ.predecessor.id,
-                             pred_succ.successor.id))
+            works.update((pred_succ.predecessor,
+                    pred_succ.successor))
 
             if pred_succ.predecessor.parent:
-                parent_ids.add(pred_succ.predecessor.parent.id)
+                parents.add(pred_succ.predecessor.parent)
 
-        res = super(PredecessorSuccessor, self).delete(ids)
+        super(PredecessorSuccessor, cls).delete(pred_succs)
 
-        for work_id in work_ids:
-            work_obj.reset_leveling(work_id)
+        for work in works:
+            work.reset_leveling()
 
-        for parent_id in parent_ids:
-            work_obj.compute_dates(parent_id)
+        for parent in parents:
+            parent.compute_dates()
 
-        return res
+    @classmethod
+    def create(cls, vlist):
+        pred_succs = super(PredecessorSuccessor, cls).create(vlist)
 
-    def create(self, values):
-        work_obj = Pool().get('project.work')
-        ps_id = super(PredecessorSuccessor, self).create(values)
+        for pred_succ in pred_succs:
+            pred_succ.predecessor.reset_leveling()
+            pred_succ.successor.reset_leveling()
 
-        pred_succ = self.browse(ps_id)
-        work_obj.reset_leveling(pred_succ.predecessor.id)
-        work_obj.reset_leveling(pred_succ.successor.id)
-
-        if pred_succ.predecessor.parent:
-            work_obj.compute_dates(pred_succ.predecessor.parent.id)
-        return id
-
-PredecessorSuccessor()
+            if pred_succ.predecessor.parent:
+                pred_succ.predecessor.parent.compute_dates()
+        return pred_succs
 
 
 class Leveling(Wizard):
     'Tasks Leveling'
-    _name = 'project_plan.work.leveling'
+    __name__ = 'project_plan.work.leveling'
+    start_state = 'leveling'
+    leveling = StateTransition()
 
-    states = {
-        'init': {
-            'result': {
-                'type': 'action',
-                'action': '_leveling',
-                'state': 'end',
-                },
-            },
-        }
-
-    def _leveling(self, data):
-        work_obj = Pool().get('project.work')
-        work_obj.create_leveling(data['id'])
-        return {}
-
-Leveling()
+    def transition_leveling(self):
+        Work = Pool().get('project.work')
+        Work(Transaction().context['active_id']).create_leveling()
+        return 'end'
